#!/usr/bin/env python

# exit codes
NOT_OK = 1
OK = 0

# modules for OPUS internal poller
import sys                  # access to command-line args
import string               # string functions
import opusutil             # utilites like PrintMsg
import re                   # regular expression handling
import traceback            # printing of stack traces

# python modules for this application
import os
from stat import *
import getopt
import glob
import time

# local modules for this application
import base_poller
import hst_archdbquery

# exceptions
class UsageError(Exception):
  pass

class HaltEvent(Exception):
  pass

class DatasetNotFoundInOSF(Exception):
  pass

class DataIdNotFoundInOSF(Exception):
  pass

class NoFilesFound(Exception):
  pass

class MissingRequiredResourceEntry(Exception):
  pass

class UnsupportedDataID(Exception):
  pass

class InputFilesError(Exception):
  pass

class FailedMatch(Exception):
  pass

class CatalogASCII(base_poller.InternalPoller):
  """
########################################################################
Class: CatalogASCII

Description:
------------
The class for interactive or pipeline CatalogASCII determination.

Members:
--------
log_dir = location of log files
arch_dbobj = database object for accessing ARCH_DB tables

Methods:
--------
usage() - prints an interactive usage message and raises an exception

History:
08/16/04 49860 MSwam   fix errant usage message caught in 15.4 testing

########################################################################
  """
  def __init__(self):
    # initialize members
    self._log_dir = None
    self._inpath = None
    self._arch_dbobj       = None
    self._arch_server      = os.environ['ARCH_SERVER']
    self._arch_db          = os.environ['ARCH_DB']

  def initialize(self):
    #
    # call superclass constructor
    base_poller.InternalPoller.__init__(self)

  #--------------------------------------------------------------------------
  def usage(self):
    opusutil.PrintMsg('F','Interactive Usage: catalog_ascii_main [-h] -d datasetname -a data_id -r resource_filespec')
    raise UsageError

  #--------------------------------------------------------------------------
  #
  # Name: interactive
  #
  # Description: Performs command line parsing and set-up for CATALOG_ASCII
  #              processing in interactive mode.  Runs the CATALOG_ASCII
  #              task for the command-line arguments given.
  #
  # Inputs: global sys.argv (command line arguments)
  #          
  # Notes:
  #        Will EXIT (status=1) if incorrect command line arguments
  #        are found.  This is a FATAL error.
  #
  # History:
  #   date    opr     who     reason
  # -------- -----  --------  --------------------------------------
  # 01/07/04 49754  MSwam     first version
  # 07/15/10 57271  MSwam     remove db retry parameter
  #
  #--------------------------------------------------------------------------
  def interactive(self):
    #
    # parse command line arguments
    try:
      # can't use "-i" (used by the OAPI)
      opts, args = getopt.getopt(sys.argv[1:], "hd:a:r:", 
                   ["help","datasetname=","data_id=",
                    "resource_filespec="])
    except getopt.GetoptError:
      # print help information and raise exception
      self.usage()

    # initialize
    dataset = None
    data_id = None

    # parse options
    for flag, value in opts:
      if flag in ("-h", "--help"):
        self.usage()
      if flag in ("-d", "--datasetname"):
        dataset = value
      if flag in ("-a", "--data_id"):
        data_id = value
      if flag in ("-r", "--resource_filespec"):
        resource_filespec = value

    # verify required args
    if (dataset == None or data_id == None):
      self.usage()

    # connect to dbs
    self._arch_dbobj = hst_archdbquery.HST_Arch_db(self._arch_server,
                                         self._arch_db)

    # set the directory for file operations to the current directory
    self._log_dir = os.getcwd() + os.sep
    self._inpath = self._log_dir

    # run the task
    status = self.do_catalog_ascii(dataset, data_id)

    # close the database connections
    self._arch_dbobj.close()

    return

  #-------------------------------------------------------------------------
  # 
  # Name: pipeline_init
  #
  # Description: Initialization for pipeline mode, including reading
  #              resource file parameters for controlling the task.
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 01/07/04 49754  MSwam     first version
  #
  #-------------------------------------------------------------------------
  def pipeline_init(self):
    # obtain resource items in a dictionary and assign to globals
    #
    try:
      resource_items = base_poller.InternalPoller.get_resource_items(self,
                            ["LOG_DIR",
                             "INPATH",
                             "ARCH_DB",
                             "ARCH_SERVER"] )
    except:
      opusutil.PrintMsg("E","failed to get required resource entry")
      raise
    #
    opusutil.PrintMsg("I","Resource items read.")
    self._log_dir = resource_items['LOG_DIR']
    self._inpath = resource_items['INPATH']
    self._arch_server = resource_items['ARCH_SERVER']
    self._arch_db = resource_items['ARCH_DB']

    # connect to the databases
    self._arch_dbobj = hst_archdbquery.HST_Arch_db(self._arch_server,
                                         self._arch_db)

  #-------------------------------------------------------------------------
  # 
  # Name: parse_sms_file
  #
  # Description: Extract values from SMS file and store them in a map.
  #
  # Input:
  #   filespec - name of sms file
  #
  # Returns:
  #   values - a map of values extracted from the SMS file,
  #                keyed on parameter name
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 01/07/04 49754  MSwam     first version
  # 09/22/04 51879  MSwam     handle SMS format more generically
  #
  #-------------------------------------------------------------------------
  def parse_sms_file(self, filespec):
    #
    values = {}
    bigstring = ""
    #
    # open the file
    fd = open(filespec, 'r')
    #
    # read lines until SMSTIME is found, concatenating them into a big string
    # (trim newlines)
    count = 0
    maxlines = 20
    line = fd.readline()
    while (string.find(line,"SMSTIME") == -1 and count < maxlines):
      bigstring = bigstring + line[:-1]
      count = count + 1
      line = fd.readline()
    #
    # add the matching SMSTIME line as well
    bigstring = bigstring + line
    fd.close() 

    #
    # CALENDAR, and PDB_ID - use regex searches
    #   where the 1st group contains ",CALENDAR(", (the paren needs escaping)
    #         the 2nd group is one or more chars, and 
    #         the 3rd group is ")"   (again the paren needs escaping)
    #  
    #   the CALENDAR value is in the 2nd group, same for PDB_ID
    #
    calendar = re.search("(,CALENDAR\()(.+?)(\))",bigstring)
    if calendar == None:
      opusutil.PrintMsg("E","Failed to determine CALENDAR for :"+bigstring)
      raise FailedMatch
    values["CALENDAR"] = calendar.group(2)
    opusutil.PrintMsg("D","calendar = "+str(values["CALENDAR"]))

    pdb_id = re.search("(,PDB_ID\()(.+?)(\))",bigstring)
    if pdb_id == None:
      opusutil.PrintMsg("E","Failed to determine PDB_ID for :"+bigstring)
      raise FailedMatch
    values["PDB_ID"] = pdb_id.group(2)
    opusutil.PrintMsg("D","pdb_id = "+str(values["PDB_ID"]))
    #
    return values

  #-------------------------------------------------------------------------
  # 
  # Name: process_sms
  #
  # Description: Extract values from SMS file and populate the archive
  #              catalog tables with them.
  #
  # Input:
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 01/07/04 49754  MSwam     first version
  # 05/06/04 49754  MSwam     use .* as wildcard, to avoid picking up FITS
  #
  #-------------------------------------------------------------------------
  def process_sms(self, dataset, data_id):
    #
    # look up the archive class for the given data_id
    archive_class = "SMS"
    #
    # locate the SMS file by globbing the inpath+dataset+wildcard
    file_count = 0
    filespec = self._inpath + os.sep + dataset + ".*"
    opusutil.PrintMsg("D","before glob,filespec="+str(filespec))
    flist = glob.glob(filespec)
    for f in flist:
      opusutil.PrintMsg("I","found "+str(f))
      #
      # parse the values needed from the file and store them in a map
      sms_values = self.parse_sms_file(f)
      #
      #
      # get file creation date as determined by
      # mtime: Time when data was last modified. Changed by the
      #  following  functions:
      #    creat(), mknod(), pipe(), utime(), write(2)
      #
      numeric_creation_date = os.stat(f)[ST_MTIME]
      creation_date = time.strftime("%b %d %Y %H:%M:%S",
                                    time.localtime(numeric_creation_date))
      opusutil.PrintMsg("I","creation_date="+str(creation_date))
      #
      # call a routine to load the archive catalog with the values 
      self._arch_dbobj.insert_sms_data(archive_class, 
                                     dataset, 
                                     creation_date,
                                     sms_values["PDB_ID"],
                                     sms_values["CALENDAR"])
      #
      file_count += 1
    #
    if file_count == 0:
      opusutil.PrintMsg("E","No files found for "+filespec)
      raise InputFilesError
    elif file_count != 1:
      opusutil.PrintMsg("E","Found multiple files for "+filespec)
      raise InputFilesError
    #
    return

  #-------------------------------------------------------------------------
  # 
  # Name: do_catalog_ascii
  #
  # Description: 
  #
  # Input:
  #   dataset
  #   data_id
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 01/07/04 49754  MSwam     first version
  #
  #-------------------------------------------------------------------------
  def do_catalog_ascii(self, dataset, data_id):
    #
    # SMS data fills its own archive table
    if data_id == "sms":
      self.process_sms(dataset, data_id)
    else:
      opusutil.PrintMsg("E","Not a supported data_id for ASCII cataloguing:"+
                        data_id)
      raise UnsupportedDataID
    #
    return 0

  #-------------------------------------------------------------------------
  # Name: processHaltEvent
  #
  # Description: Handle a HALT event (override base class).
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 09/22/04 51879  MSwam     use base poller class
  #
  #-------------------------------------------------------------------------
  def processHaltEvent(self, event):
    #
    opusutil.PrintMsg("I","Received HALT event. Stopping.")
    self._arch_dbobj.close()
    raise HaltEvent

  #-------------------------------------------------------------------------
  # Name: processOsfEvent
  #
  # Description: Handle OSF events in pipeline mode.
  #
  # History:
  # -------- -----  --------  --------------------------------------
  # 04/01/04 49754  MSwam     first version
  # 05/11/04 49754  MSwam     fix success trigger name
  # 08/06/04 49860  Sontag    Changed an IDL field name
  # 09/22/04 51879  MSwam     Use base_poller class
  # 10/29/04 52310  MSwam     remove db closes (only HALT event should)
  #
  #-------------------------------------------------------------------------
  def processOsfEvent(self, event):
        # get trigger info
        dataset, data_id = self.getDatasetDataidFromOsfEvent(event)

        try:
          opusutil.OpenTrl(self._log_dir + os.sep + dataset + "_" +
                           data_id + ".cat")
        except:
          # any exception marks the OSF with OSF_FAIL
          traceback.print_exc()
          #
          # an error occurred during trailer opening, set the OSF to failed
          try:
            opusutil.PrintMsg("E","closing event with OSF_FAIL")
            self._papp.closeEvent("OSF_FAIL", event.eid)
          except:
            traceback.print_exc()
            opusutil.PrintMsg("E","Failed to close OSF_FAIL event")
          return

        opusutil.PrintMsg("I","----- CAT_ASCII started for: "+dataset+"_"+
                          data_id+" -----")
        try:
          mode = self.do_catalog_ascii(dataset, data_id)
        except:
          # any other exception marks the OSF with OSF_FAIL
          traceback.print_exc()
          #
          # an error occurred during processing, set the OSF
          try:
            opusutil.PrintMsg("E","closing event with OSF_FAIL")
            self._papp.closeEvent("OSF_FAIL", event.eid)
          except:
            traceback.print_exc()
            opusutil.PrintMsg("E","Failed to close OSF_FAIL event")
          opusutil.PrintMsg("I","----- CAT_ASCII ended for: "+dataset+"_"+
                            data_id+" -----")
          opusutil.CloseTrl()
          return

        try:
          self._papp.closeEvent("OSF_NORMAL", event.eid)
        except:
            traceback.print_exc()
            opusutil.PrintMsg("E","Failed to close OSF_NORMAL event")
        opusutil.PrintMsg("I","----- CAT_ASCII ended for: "+dataset+"_"+
                          data_id+" -----")
        opusutil.CloseTrl()

#-------------------------------------------------------------------------
# The main routine. Handles interactive and pipeline mode.
#
# History:
# -------- -----  --------  --------------------------------------
# 09/22/04 51879  MSwam     first version
#
#-------------------------------------------------------------------------
#
# initialize
try:
  poller = CatalogASCII()
  poller.initialize()
  poller.pipeline_init()
#
except base_poller.InteractiveMode:
  opusutil.PrintMsg("I","Running command-line mode.")
  try:
    poller.interactive()
  except UsageError:
    sys.exit(NOT_OK)
  except:
    traceback.print_exc()
    opusutil.PrintMsg("F","Command-line mode Failed.")
    sys.exit(NOT_OK)
  sys.exit(OK)
#
except:
  traceback.print_exc()
  opusutil.PrintMsg("E","initialization failed")
  sys.exit(NOT_OK)

opusutil.PrintMsg("I","Pipeline set-up complete.")
try:
  poller.run()
except HaltEvent:
  sys.exit(OK)
except:
  traceback.print_exc()
  opusutil.PrintMsg("E","pipeline_process failed")
  sys.exit(NOT_OK)
